/*****************************************************************************
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: phm2sdf.c,v 1.5 2000/04/30 19:17:54 kevin Exp $
**  $Log: phm2sdf.c,v $
**  Revision 1.5  2000/04/30 19:17:54  kevin
**  *** empty log message ***
**
**  Revision 1.4  2000/04/30 04:06:13  kevin
**  Update Raysum i/o routines
**  Fix MPI bug in ctrec (scatter_raysum) that referenced rs_global
**
**  Revision 1.3  2000/04/28 18:19:01  kevin
**  removed unused files
**
**  Revision 1.2  2000/04/28 13:50:45  kevin
**  Removed Makefile Makefile.in that are automatically generated by autoconf
**
**  Revision 1.1.1.1  2000/04/28 13:02:44  kevin
**  Initial CVS import for first public release
**
**
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/
/* FILE
 *   phm2sdf.c			Generate a SDF image from a phantom
 *
 * HISTORY
 *   1984 - Final DOS verion
 *   1999 - First UNIX version
 */

#include "ct.h"

#define O_PHANTOM       1
#define O_DESC          2
#define O_NSAMPLE       3
#define O_FILTER        4
#define O_TRACE         5
#define O_VERBOSE       6
#define O_HELP          7
#define O_PICFILE       8
#define O_FILTER_DOMAIN 9
#define O_FILTER_BW     10
#define O_FILTER_PARAM  11
#define O_DEBUG         12
#define O_VERSION       13

static struct option my_options[] = 
{
  {"phantom", 1, 0, O_PHANTOM},
  {"picfile", 1, 0, O_PICFILE},
  {"desc", 1, 0, O_DESC},
  {"nsample", 1, 0, O_NSAMPLE},
  {"filter", 1, 0, O_FILTER},
  {"filter-domain", 1, 0, O_FILTER_DOMAIN},
  {"filter-bw", 1, 0, O_FILTER_BW},
  {"filter-param", 1, 0, O_FILTER_PARAM},
  {"trace", 1, 0, O_TRACE},
  {"verbose", 0, 0, O_VERBOSE},
  {"debug", 0, 0, O_DEBUG},
  {"help", 0, 0, O_HELP},
  {"version", 0, 0, O_VERSION},
  {0, 0, 0, 0}
};

void 
usage (const char *program)
{
  fprintf(stdout,"usage: %s outfile nx ny [--phantom phantom-name] [--picfile filename] [--filter filter-name] [OPTIONS]\n",kbasename(program)); 
  fprintf(stdout,"Generate phantom image from a predefined --phantom or a --picfile\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"     outfile         Name of output file for image\n");
  fprintf(stdout,"     nx              Number of pixels X-axis\n");
  fprintf(stdout,"     ny              Number of pixels Y-axis\n");
  fprintf(stdout,"     --phantom       Phantom to use for projection\n");
  fprintf(stdout,"        herman       Herman head phantom\n");
  fprintf(stdout,"        rowland      Rowland head phantom\n");
  fprintf(stdout,"        browland     Bordered Rowland head phantom\n");
  fprintf(stdout,"        unitpulse    Unit pulse phantom\n");
  fprintf(stdout,"     --picfile       Generate Phantom from picture file\n");
  fprintf(stdout,"     --filter        Generate Phantom from a filter function\n");
  fprintf(stdout,"       abs_bandlimit Abs * Bandlimiting\n");
  fprintf(stdout,"       abs_sinc      Abs * Sinc\n");
  fprintf(stdout,"       abs_cos       Abs * Cosine\n");
  fprintf(stdout,"       abs_hamming   Abs * Hamming\n");
  fprintf(stdout,"       shepp         Shepp-Logan\n");
  fprintf(stdout,"       bandlimit     Bandlimiting\n");
  fprintf(stdout,"       sinc          Sinc\n");
  fprintf(stdout,"       cos           Cosine\n");
  fprintf(stdout,"       triangle      Triangle\n");
  fprintf(stdout,"       hamming       Hamming\n");
  fprintf(stdout,"     --filter-param  Alpha level for Hamming filter\n");
  fprintf(stdout,"     --filter-domain Set domain of filter\n");
  fprintf(stdout,"         spatial     Spatial domain (default)\n");
  fprintf(stdout,"         freq        Frequency domain\n");
  fprintf(stdout,"     --filter-bw     Filter bandwidth (default = 1)\n");
  fprintf(stdout,"     --desc          Description of raysum\n");
  fprintf(stdout,"     --nsample       Number of samples per axis per pixel (default = 1)\n");
  fprintf(stdout,"     --trace         Trace level to use\n");
  fprintf(stdout,"        none         No tracing (default)\n");
  fprintf(stdout,"        text         Trace text level\n");
  fprintf(stdout,"        pic          Trace picture\n");
  fprintf(stdout,"        rays         Trace rays\n");
  fprintf(stdout,"        plot         Trace plot\n");
  fprintf(stdout,"        clipping     Trace clipping\n");
  fprintf(stdout,"     --debug         Debug mode\n");
  fprintf(stdout,"     --verbose       Verbose mode\n");
  fprintf(stdout,"     --version       Print version\n");
  fprintf(stdout,"     --help          Print this help message\n");
  exit(1);
}

#ifdef MPI_CT
void mpi_gather_image (IMAGE *im_global, IMAGE *im_local, const int opt_debug);
#endif

int 
main (const int argc, char *const argv[])
{
  IMAGE *im_global = NULL;
  PICTURE *pic = NULL;
  int opt_nx = 0, opt_ny = 0;
  int opt_nsample = 1;
  int opt_picnum = -1;
  int opt_filter = -1;
  int opt_filter_domain = D_SPATIAL;
  char *opt_outfile = NULL;
  int opt_debug = 0;
  char str[256];
  char opt_desc[256], opt_picfilename[256];
  char *endstr, *endptr;
  double opt_filter_param = 1;
  double opt_filter_bw = 1.;
  int opt_trace = TRACE_NONE;
  int opt_verbose = 0;
  double time_start=0, time_end=0;
#ifdef MPI_CT
  IMAGE *im_local = NULL;
  int mpi_argc = argc;
  char **mpi_argv = (char **) argv;
  double mpi_t1, mpi_t2, mpi_t, mpi_t_g;

  MPI_Init(&mpi_argc, &mpi_argv);
  MPI_Comm_dup (MPI_COMM_WORLD, &mpi_ct.comm);
  MPI_Comm_size(mpi_ct.comm, &mpi_ct.nproc);
  MPI_Comm_rank(mpi_ct.comm, &mpi_ct.my_rank);

  if (mpi_ct.nproc > MPI_MAX_PROCESS) {
    sys_error(ERR_FATAL, "Number of mpi processes (%d) exceeds max processes (%d)",
	      mpi_ct.nproc, MPI_MAX_PROCESS);
    exit(1);
  }
#endif

#ifdef MPI_CT
  time_start = MPI_Wtime();
#else
  time_start = td_current_sec();
#endif
  
#ifdef MPI_CT
  if (mpi_ct.my_rank == 0) {
#endif

    strcpy(opt_desc, "");
    strcpy(opt_picfilename, "");
    while (1) {
      int c = getopt_long(argc, argv, "", my_options, NULL);
      char *endptr = NULL;
      char *endstr;
      
      if (c == -1)
	break;
      
      switch (c) {
      case O_PHANTOM:
	opt_picnum = opt_set_picture(optarg, argv[0]);
	break;
      case O_PICFILE:
	strncpy(opt_picfilename, optarg, sizeof(opt_picfilename));
	pic = create_pic_from_file(opt_picfilename);
#ifdef MPI_CT
	if (mpi_ct.my_rank == 0) 
	  fprintf(stderr, "Can't use picture from file in MPI mode\n");
	exit(1);
#endif
	break;
      case O_VERBOSE:
	opt_verbose = 1;
	break;
      case O_DEBUG:
	opt_debug = 1;
	break;
      case O_TRACE:
	opt_trace = opt_set_trace(optarg, argv[0]);
	break;
      case O_FILTER:
	opt_filter = opt_set_filter(optarg, argv[0]);
	break;
      case O_FILTER_DOMAIN:
	opt_filter_domain = opt_set_filter_domain(optarg, argv[0]);
	break;
      case O_DESC:
	strncpy(opt_desc, optarg, sizeof(opt_desc));
	break;
      case O_FILTER_BW:
	opt_filter_bw = strtod(optarg, &endptr);
	endstr = optarg + strlen(optarg);
	if (endptr != endstr) {
	  fprintf(stderr,"Error setting --filter-bw to %s\n", optarg);
	  usage(argv[0]);
	  exit(1);
	}
	break;
      case O_FILTER_PARAM:
	opt_filter_param = strtod(optarg, &endptr);
	endstr = optarg + strlen(optarg);
	if (endptr != endstr) {
	  fprintf(stderr,"Error setting --filter-param to %s\n", optarg);
	  usage(argv[0]);
	  exit(1);
	}
	break;
      case O_NSAMPLE:
	opt_nsample = strtol(optarg, &endptr, 10);
	endstr = optarg + strlen(optarg);
	if (endptr != endstr) {
	  fprintf(stderr,"Error setting --nsample to %s\n", optarg);
	  usage(argv[0]);
	  exit(1);
	}
	break;
        case O_VERSION:
#ifdef VERSION
	  fprintf(stdout, "Version %s\n", VERSION);
#else
	  fprintf(stderr, "Unknown version number");
#endif
      case O_HELP:
      case '?':
	usage(argv[0]);
	exit(0);
      default:
	usage(argv[0]);
	exit(1);
      }
    }
    
    if (pic == NULL && opt_picnum == -1 && opt_filter == -1) {
      fprintf(stderr, "No phantom defined\n");
      usage(argv[0]);
      exit(1);
    }

    if (opt_trace >= TRACE_PIC)
      show_pic(pic);

    if (optind + 3 != argc) {
      usage(argv[0]);
      exit(1);
    }
    opt_outfile = argv[optind];
    opt_nx = strtol(argv[optind+1], &endptr, 10);
    endstr = argv[optind+1] + strlen(argv[optind+1]);
    if (endptr != endstr) {
      fprintf(stderr,"Error setting nx to %s\n", argv[optind+1]);
      usage(argv[0]);
      exit(1);
    }
    opt_ny = strtol(argv[optind+2], &endptr, 10);
    endstr = argv[optind+2] + strlen(argv[optind+2]);
    if (endptr != endstr) {
      fprintf(stderr,"Error setting ny to %s\n", argv[optind+2]);
      usage(argv[0]);
      exit(1);
    }
    
    snprintf(str, sizeof(str), "nx=%d, ny=%d, nsample=%d, ", opt_nx, opt_ny, opt_nsample);
    if (opt_picfilename[0]) {
      strncat(str, "phantom=", sizeof(str));
      strncat(str, opt_picfilename, sizeof(str));
    }
    else if (opt_picnum != -1) {
      strncat(str, "phantom=", sizeof(str));
      strncat(str, name_of_picture(opt_picnum), sizeof(str));
    }
    else if (opt_filter != -1) {
      strncat(str, "filter=", sizeof(str));
      strncat(str, name_of_filter(opt_filter), sizeof(str));
      strncat(str, " - ", sizeof(str));
      strncat(str, name_of_filter_domain(opt_filter_domain), sizeof(str));
    }
    if (opt_desc[0]) {
      strncat(str, ": ", sizeof(str));
      strncat(str, opt_desc, sizeof(str));
    }
    strncpy(opt_desc, str, sizeof(opt_desc));
    
    if (opt_verbose)
      printf("Compile Phantom to Image\n\n");
#ifdef MPI_CT
  }
#endif

#ifdef MPI_CT
  mpi_t1 = MPI_Wtime();
  MPI_Bcast(&opt_verbose, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_debug, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_trace, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_nx, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_ny, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_nsample, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_picnum, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_filter, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_filter_domain, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_filter_param, 1, MPI_DOUBLE, 0, mpi_ct.comm);
  MPI_Bcast(&opt_filter_bw, 1, MPI_DOUBLE, 0, mpi_ct.comm);

  if (opt_verbose) {
    mpi_t2 = MPI_Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    MPI_Reduce(&mpi_t, &mpi_t_g, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_ct.comm);
    if (mpi_ct.my_rank == 0)
      printf("Time to Bcast vars = %f secs, Max time = %f\n", mpi_t, mpi_t_g);
  }

  mpi_ct_calc_work_units(opt_nx);

  if (mpi_ct.my_rank == 0)
    im_global = image_create (opt_outfile, opt_nx, opt_ny);

  im_local = image_create(NULL, opt_nx, opt_ny);
#else
  im_global = image_create (opt_outfile, opt_nx, opt_ny);
#endif
  
  if (opt_picnum > 0)
    pic = create_pic(opt_picnum);
#ifdef MPI_CT
  else {
    if (mpi_ct.my_rank == 0)
      fprintf(stderr, "picnum < 0\n");
    exit(1);
  }
#endif

#ifdef MPI_CT
  if (pic->type == P_UNIT_PULSE) {
    if (mpi_ct.my_rank == 0) {
      im_global->v[opt_nx/2][opt_ny/2] = 1.;
      im_global->calctime = 0;
    }
  } else if (opt_filter != -1) {
    if (mpi_ct.my_rank == 0) {
      image_filter_init (im_global, opt_filter_domain, opt_filter_bw, opt_filter, opt_filter_param, opt_trace);
      im_global->calctime = 0;
    }
  } else {
    if (opt_verbose)
      mpi_t1 = MPI_Wtime();
    pic_to_image (pic, im_local, mpi_ct.start_work_unit[mpi_ct.my_rank],
		  mpi_ct.local_work_units[mpi_ct.my_rank], opt_nsample, opt_trace);
    if (opt_verbose) {
      mpi_t2 = MPI_Wtime();
      mpi_t = mpi_t2 - mpi_t1;
      MPI_Reduce(&mpi_t, &mpi_t_g, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_ct.comm);
      if (mpi_ct.my_rank == 0)
	printf("Time to compile pic = %f secs, Max time = %f\n", mpi_t, mpi_t_g);
    }
    mpi_gather_image(im_global, im_local, opt_debug);
  }
#else
  if (pic->type == P_UNIT_PULSE) {
    im_global->v[opt_nx/2][opt_ny/2] = 1.;
    im_global->calctime = 0;
  } else if (opt_filter != -1) {
    image_filter_init (im_global, opt_filter_domain, opt_filter_bw, opt_filter, opt_filter_param, opt_trace);
    im_global->calctime = 0;
  } else {
    pic_to_image (pic, im_global, 0, opt_nx, opt_nsample, opt_trace);
  }
#endif

#ifdef MPI_CT
  time_end = MPI_Wtime();
  if (mpi_ct.my_rank == 0) {
    im_global->calctime = time_end - time_start;
    strncpy (im_global->remark, opt_desc, sizeof(im_global->remark));
  }

  if (mpi_ct.my_rank == 0) {
    image_save (im_global);
    if (opt_verbose)
      fprintf (stdout, "Time to compile image = %.2f sec\n\n", im_global->calctime);
  }
#else
  time_end = td_current_sec();
  im_global->calctime = time_end - time_start;
  strncpy (im_global->remark, opt_desc, sizeof(im_global->remark));
  image_save (im_global);

  if (opt_trace >= TRACE_PIC)
    {
      crt_set_cpos (1, 1);
    }
  
  if (opt_verbose)
    fprintf (stdout, "Time to compile image = %.2f sec\n\n", im_global->calctime);
#endif

  if (opt_trace >= TRACE_PIC)
    {
      double dmin, dmax;
      int nx, ny;

      printf ("Enter size of cell (nx, ny): ");
      scanf ("%d %d", &nx, &ny);
      printf ("Enter minimum and maximum densities (min, max): ");
      scanf ("%lf %lf", &dmin, &dmax);
      //      image_paint (CRTDEV, im_global, 0, 0, nx, ny, dmin, dmax, 0);
      WAITKEY();

      crt_put_stra ("Finished", 8 + C_WHITE);
      crt_set_cpos (1, 2);
    }

  return (0);
}



#ifdef MPI_CT
void mpi_gather_image (IMAGE *im_global, IMAGE *im_local, const int opt_debug)
{
  int iproc;
  int end_work_unit;
  int iw;

  end_work_unit = mpi_ct.local_work_units[mpi_ct.my_rank] - 1;
  for (iw = 0; iw <= end_work_unit; iw++) {
    MPI_Send(im_local->v[iw], im_local->ny, MPI_FLOAT, 0, 0, mpi_ct.comm);
  }

  if (mpi_ct.my_rank == 0) {
    for (iproc = 0; iproc < mpi_ct.nproc; iproc++) {
      end_work_unit = mpi_ct.start_work_unit[iproc] 
	+ mpi_ct.local_work_units[iproc] 
	- 1;

      if (opt_debug) {
	fprintf(stdout, "Recv rs data from process %d (%d-%d)\n", iproc,
		mpi_ct.start_work_unit[iproc], end_work_unit);
	fflush(stdout);
      }


      for (iw = mpi_ct.start_work_unit[iproc]; iw <= end_work_unit; iw++) {
	MPI_Status status;

	MPI_Recv(im_global->v[iw], im_global->ny, MPI_FLOAT, iproc, 0, mpi_ct.comm, &status);
      }
    }
  }

}
#endif
